/**
 * GenerationMethods.cpp - Bitcoin address generator
 *
 * Copyright (C)2015 The bitgen developers <bitgen@openmailbox.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
 
/**
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE. 
*/
 

#include "GenerationMethods.h"
#include "GenerationMethodsExtended.h"
#include "GenerationMethodsElectrum.h"

#include "GenerationParameters.h"
#include "commandLine.h"
#include "util.h"
#include "random.h"
#include "privKey.h"
#include "bitcoinKeyPair.h"
#include "mnemonic.h"
#include "SaveFolderHandler.h"
#include "txtGeneration.h"
#include "qrenc.h"
#include "randomUtils.h"
#include "FileUtil.h"
#include "currencySpecific.h"
#include "keyDerivationFunction.h"
#include "bitgenVersion.h"
#include "printPS.h"
#include "ExtendedKey.h"
#include "hmac.h"
#include "bitgenConfig.h"
#include "StaticXor.h"
#include "pubXmrAddress.h"
#include "OtpFolderHandler.h"
#include "Directory.h"
#include "InformationList.h"


#include <iostream>
#include <list>
#include <stdlib.h>
#include <algorithm> 
#include <iterator>
#include <map>
#include <sstream>







void GenerationMethodsOtp::saveTheOTPPublicFilesInFolders(const AddressGenerationInfo& allInfo, const std::string& pubPath)
{		
	const std::string filenameBase = getCurrencyShortName() + "_address_" + intToString(allInfo.getNewNum()) + "_pad";
	
	if(allInfo.params.getGenerateTXT())
	{		
		savePubOtpTxtFiles(filenameBase, allInfo.pubaddr, pubPath, allInfo);
	}
	
	if(allInfo.params.getGeneratePS())
	{		
		const Language lang = getLanguage();		
		const PostscriptGenerator postscriptGenerator(allInfo, versionNumber, applicationName, lang);
		
		const PublicPostscriptInfo pubInfo(allInfo);
		
		const std::string privaddr("DummyPrivAddr");
		const std::list<ExtraInfo> generationExtra;
		const std::string mnemonicDummy;
		
		
		const PrivatePostscriptInfo privInfo(allInfo);
		
		const std::string psinfoPublic = postscriptGenerator.postscriptPagePublic(pubInfo);
		
		const std::string pubFilename = filenameBase + "_" + allInfo.pubaddr.substr(0, 8) + "_share.ps";
		
		const std::string fullPubPath = (pubPath == ".") ? pubFilename : (pubPath + "/" + pubFilename);	
		std::cout << "Saving publfile: " << fullPubPath << std::endl;
		fileWrite(fullPubPath, psinfoPublic);
	}	
}



void GenerationMethodsOtp::saveAllFilesOtp(const AddressGenerationInfo& allInfo, const bool useConf)
{	
	std::string privPath(".");
	std::string pubPath(".");

	if(allInfo.params.getUseFolder())
	{	
		const bool doCreateFolder = !allInfo.getIsPad(); //isPad;
		std::list<std::string> confFolders;
		confFolders.push_back("");
		const std::pair<std::string, std::vector<std::string> > folders = SaveFolderHandler::createNewFoldersOtp(allInfo.getNewNum(), confFolders, "", doCreateFolder, true);
		
		privPath = folders.second.front();
		pubPath = folders.first;
	}
	
	GenerationMethods::saveTheFilesInFolders(allInfo, privPath, pubPath);	
}




void savePadFile(const AddressGenerationInfo allInfo, const BigInt<256>& pad, const int numPad, const int totalPads, const std::string& privPath, const std::string& privFilename)
{	
	if(allInfo.params.getGenerateTXT())
	{	
		const std::string txtFilename = privFilename + ".txt";
		//std::cout << "Will save txt" << std::endl;
		saveOtpFile(allInfo.pubaddr, pad.getBinaryLong(), numPad, totalPads, allInfo.getTime(), privPath, versionNumber, 
			applicationName, allInfo.getGenerationExtra(), allInfo.getAccountName(), allInfo.getGenerator(), allInfo.getStaticXor(), allInfo.getNewNum(), 
			txtFilename);
	}
	
	
	if(allInfo.params.getGeneratePS())
	{		
		const Language lang = getLanguage();		
		const PostscriptGenerator postscriptGenerator(allInfo, versionNumber, applicationName, lang);
		
		//Pubaddr is not displayed, instead a hash of the pubaddr is used.
		const std::string emptyPubAddr;
		
		const bool useHashOfTimeInsteadOfAddress(true);		//Use hash of time instead of address?
		const PublicPostscriptInfo pubInfo(allInfo, useHashOfTimeInsteadOfAddress);
		
		const PrivatePostscriptInfo privInfo(allInfo, true, (numPad + 1), totalPads);		
		
		const PsInfoFile psinfo = postscriptGenerator.generate(privInfo, pubInfo, true);
			
		const std::string privFilenamePs = privFilename + ".ps";
		const std::string fullPrivPath = (privPath == ".") ? privFilenamePs : (privPath + "/" + privFilenamePs);
	
		std::cout << "Saving priv pad: " << withoutLeadingDot(fullPrivPath) << std::endl;
		fileWrite(fullPrivPath, psinfo.privatePs);		
	}
}



std::string parse256BitHexString(const std::string& hexVal)
{
	const int numBytes = 32;
	
	if(hexVal.size() != (numBytes * 2))
	{
		std::cerr << std::endl;
		std::cerr << "Incorrect amount of data given. Got " << hexVal.size() << " hex digits, want " << (numBytes * 2) << " hex digits" << std::endl;
		throw std::string("Not enough data");
	}
		
	const std::string binStr = convertFromHex(hexVal);
	return binStr;
}

std::list<BigInt<256> > getPadsFromCommandline(const std::vector<std::string>& args, bool& useStaticXorIfPresent)
{
	std::list<BigInt<256> > pads;
	
	const int numArgs = args.size() - 1;
	
	for(int i = 0 ; i < numArgs ; i++)
	{
		const std::string padHex = args[i + 1];
		
		if(padHex == "nostatic")
		{
			useStaticXorIfPresent = false;
			return pads;
		}
	
		const std::string padBin = parse256BitHexString(padHex);	
	
		const BigInt<256> padInt(padBin,'c');
		//std::cout << "FOUND PAD: " << hexString(padInt.getBinaryLong()) << std::endl;
		pads.push_back(padInt);
	}

	return pads;
}




std::list<BigInt<256> > getPadsWithNum(const int addressNum)
{
	std::list<BigInt<256>> ret;
	
	std::string publicFolder;
	const std::vector<std::string> folders = BitgenConfig::getPublicAndOtpFoldersVector(publicFolder);	
	
	for(int i = 0 ; i < folders.size() ; i++)
	{
		const std::string folder = folders[i];		
		
		const int padOneIndex = i + 1;
		
		const std::string totFolder = OtpFolderHandler::getTotalPrivFolder(folder, padOneIndex, addressNum);		
		const EntryInfo info = OtpFolderHandler::readEntryInfo(totFolder, addressNum, padOneIndex);
		
		const std::string padBin = parse256BitHexString(info.pad);		
		const BigInt<256> padInt(padBin,'c');				
		ret.push_back(padInt);
	}

	return ret;
}




std::list<BigInt<256> > getPadsFromArgs(const std::vector<std::string>& args, bool& useStaticXorIfPresent, GenerationParameters& params, int& addressNum)
{
	useStaticXorIfPresent = true;
	
	if(args.size() == 2)
	{
		
		const std::string addressNumStr = args[1];
		addressNum = atoi(addressNumStr.c_str());
		
		std::string accountText("From OTP ");
		accountText += intToString(addressNum);
		accountText += " do not use";
				
		params.setAccountName(accountText);
				
		const std::list<BigInt<256> > padList = getPadsWithNum(addressNum);		
		return padList;
	}
	else
	{
		if(args.size() < 3)
		{
			std::cout << "Usage: bitgen otp OTP1 ... OTPn" << std::endl;
			throw std::string("Incorrect usage");
		}

		params.setAccountName("From OTP do not use");
		
		const std::list<BigInt<256> > padList = getPadsFromCommandline(args, useStaticXorIfPresent);
		return padList;
	}
}




void GenerationMethodsOtp::generateOtpTotal(const std::vector<std::string>& args, const std::string& thetime)
{
	std::cout << "Generating OTP total" << std::endl;

	GenerationParameters params;	

	int addressNum(0);
	bool useStaticXorIfPresent(true);
	const std::list<BigInt<256> > padList = getPadsFromArgs(args, useStaticXorIfPresent, params, addressNum);

	const int numPads = padList.size();
	
	std::cout << "Num pads: " << numPads << std::endl;

	const BigInt<256> staticXor = BitgenConfig::getStaticXor(useStaticXorIfPresent);
	
	const BigInt<256> zero;
	if(staticXor != zero)
	{
		const std::string hash = doubleSha256FourBytes(staticXor);		
		std::cout << "Using static pad: " <<  hexString(staticXor.getBinaryLong()) << " ";
		std::cout << "(" << hexString(hash) << ")" << std::endl;	
	}
	

	BigInt<256> privKey;
	
	for(std::list<BigInt<256> >::const_iterator it = padList.begin() ;
		it != padList.end() ;
		++it)
	{		
		const BigInt<256> padInt = (*it);
		
		std::cout << "PAD: " << hexString(padInt.getBinaryLong()) << std::endl;

		privKey = padInt ^ privKey;
	}
		
	privKey = staticXor ^ privKey;

	
	std::cout << std::endl;
	std::cout << "Privkey: " << hexString(privKey.getBinaryLong()) << std::endl;
		
	const PrivKey myPrivKey(privKey, true);
		
	const std::string hexStr = hexString(privKey.getBinaryLong());

	std::list<ExtraInfo> generationExtra;
	
	const ExtraInfo extraInfo("Raw hex", hexStr);
	generationExtra.push_back(extraInfo);
	
	BitcoinEllipticCurve ec;

	const Wif wif = myPrivKey.getWif();
	const PubKey pubKey = myPrivKey.getPubkey(ec);
	const PubAddress pubAddr = pubKey.getAddress();
	
	if(usesCashAddress())
	{
		std::cout << "PUB : " << pubAddr.toCashAddrString() << std::endl;				
	}
	else
	{
		std::cout << "PUB : " << pubAddr << std::endl;		
	}
	
	std::cout << "PRIV: " << wif.toString() << std::endl;
		
	const Mnemonic mnemonicObj(privKey);

	const std::string generator = "OTP with " + intToString(numPads) + " parts" + ((staticXor != zero) ? " and staticxor" : "");
		
	std::cout << std::endl;
	
	const GenerationInfoBase infoBase(thetime, getNodeName(), applicationName, versionNumber);		
	const GenerationInfoCore infoCore(infoBase, addressNum, params);
	const GenerationInfo info(infoCore, generationExtra, generator);
	const AddressGenerationInfo allInfo(pubAddr.toString(), wif.toString(), "Wif", info);
	
	GenerationMethodsOtp::saveAllFilesOtp(allInfo);
}


BigInt<256> getPartHash(const std::string& part)
{
	const std::string hash = sha256(part);

	const BigInt<256> ret(hash, 'c');
	
	return ret;	
}

const BigInt<256> getPartDice(const std::string& part)
{
	const BigInt<256> ret = getRandomNumFrom6SideDice(part);
		
	return ret;		
}




std::list<BigInt<256> > generatePadRandom(const int totalPads, const std::string& method, const std::string& extraVal)
{
	std::list<BigInt<256> > pads;
	
	if(method == "random")
	{	
		for(int i = 0 ; i < totalPads ; i++)
		{
			const BigInt<256> pad = getRandomNumber(false);
			pads.push_back(pad);
		}
	}
	else if(method == "hash")
	{
		const int minLen = totalPads * 50;
		if(extraVal.size() < minLen)
		{
			std::cout << "Needs at least " << minLen << " character hash, got: " << extraVal.size() << std::endl;
			throw std::string("Hash too short");
		}
		
		const std::list<std::string> parts = splitHashStr(totalPads, extraVal);

		for(std::list<std::string>::const_iterator it = parts.begin() ;
			it != parts.end() ;
			++it)
		{
			const std::string& part = (*it);
			std::cout << "Hash part: " << part << std::endl;
			const BigInt<256> pad = getPartHash(part);
			pads.push_back(pad);
			
		}
	}
	else if(method == "dice")
	{
		const int neededLen = totalPads * 100;
		if(extraVal.size() != neededLen)
		{
			std::cout << "Num dice should be " << neededLen << ", got: " << extraVal.size() << std::endl;
			throw std::string("Incorrect num dice");
		}
		
		const std::list<std::string> parts = splitHashStr(totalPads, extraVal);	//Todo: Rename

		for(std::list<std::string>::const_iterator it = parts.begin() ;
			it != parts.end() ;
			++it)
		{
			const std::string& dicePart = (*it);
			std::cout << "Dice part: " << dicePart << std::endl;
			const BigInt<256> pad = getPartDice(dicePart);
			pads.push_back(pad);
			
		}
	}
	else if(method == "hex")
	{
		const int neededLen = totalPads * 64;
		if(extraVal.size() != neededLen)
		{
			std::cout << "Num hex digits should be " << neededLen << ", got: " << extraVal.size() << std::endl;
			throw std::string("Incorrect num hex digits");
		}
		
		const std::list<std::string> parts = splitHashStr(totalPads, extraVal);	//Todo: Rename

		for(std::list<std::string>::const_iterator it = parts.begin() ;
			it != parts.end() ;
			++it)
		{
			const std::string& hexPart = (*it);
			//std::cout << "Hex part: " << hexPart << std::endl;
			const std::string binStr = convertFromHex(hexPart);
		
			const std::string hexValUpper = hexString(binStr);
		
		
			const BigInt<256> pad = BigInt<256>::fromHex(hexValUpper);
			//std::cout << "PrivData: " << pad << std::endl;
			
			pads.push_back(pad);
			
		}
	}
	else
	{
		std::cout << "Internal error, no such method: " << method << std::endl;
		throw std::string("No such method");
	}


	return pads;
}






void GenerationMethodsOtp::savePublicPadPart(const AddressGenerationInfo& allInfo, const std::string& pubPath)
{		
	GenerationMethodsOtp::saveTheOTPPublicFilesInFolders(allInfo, pubPath); 
}






void GenerationMethodsOtp::savePrivatePadPart(const GenerationInfo& info, const std::list<BigInt<256> >& pads, 
	const std::list<std::string>& hashParts, std::vector<std::string>& privatePaths, const int totalPads)
{
	std::list<std::string>::const_iterator jt = hashParts.begin();
	
	
	int thisNum(0);
	for(std::list<BigInt<256> >::const_iterator it = pads.begin() ;
		it != pads.end() ;
		++it)
	{
		//std::cout << "Next pad" << std::endl;
		const BigInt<256>& pad = (*it);
		std::list<ExtraInfo> padGenerationExtra;
		
		if(info.getMethod() == "hash")
		{
			const std::string& hashVal = (*jt);
			++jt;
			const ExtraInfo extraInfoNum("Hashval", hashVal);	//Todo: Needed?
			padGenerationExtra.push_back(extraInfoNum);
		}

		const AddressGenerationInfo allInfo("", hexString(pad.getBinaryLong()), "Pad", info);

		const std::string privPath = privatePaths[thisNum];

		const std::string privFilename = getCurrencyShortName() + "_address_"+ intToString(info.getNewNum())+ 
			"_private_pad" + intToString(thisNum + 1);
		
		savePadFile(allInfo, pad,  thisNum, totalPads, privPath, privFilename);
		
		thisNum++;
	}
}










bool isPadTxtFile(const std::string& str, const int num, const int padnum)
{
	const std::string start = getCurrencyShortName() + "_address_" + intToString(num) + "_private";
		
	if(!startsWith(str, start))
	{
		return false;
	}

	const std::string end = "pad" + intToString(padnum) + ".txt";

	if(!endsWith(str, end))
	{
		return false;
	}

	return true;
}


std::string getPadFilename(const std::string& dir, const int num, const int padnum)
{
	Directory directory(dir);
	
	const std::set<std::string> files = directory.getEntries();
		
	for(const std::string& file : files)
	{
		if(!isPadTxtFile(file, num, padnum))
		{
			continue;
		}
		
		return file;
	}
	
	throw std::string("pad file not found");
	
	return "";
}




//-1 means public info
int parseNumpad(const std::string& numPadStr)
{	
	if(numPadStr == "pub")
	{
		//std::cout << "Generating public info list" << std::endl;
		return -1;
	}
	else
	{
		const int numPad = atoi(numPadStr.c_str());

		if(numPad == 0)
		{
			std::string errStr("Incorrect pad number: ");
			errStr += numPadStr;
			throw errStr;
		}
		
		return numPad;
	}	
}


int parsePubPriv(const std::string& numPadStr)
{	
	if(numPadStr == "pub")
	{
		return -1;
	}
	else if(numPadStr == "priv")
	{
		return -2;
	}	
	else
	{
		throw std::string("Command line error, expecting pub or priv");
		return 0;
	}	
}



std::string createEmptyFileContent(const int addressNum, const std::string thetime)
{
	std::string contents = "Empty " + getCurrencyShortName() + " address: " + intToString(addressNum) + "\n";
	contents += "Date: ";
	contents += thetime;

	return contents;
}



void setAsEmptyAmount(const std::string& pubFolder, const int addressNum, const std::string thetime)
{
	const std::string pubEmptyFile = OtpFolderHandler::getPubEmptyFile(pubFolder, addressNum);

	std::cout << "Writing file: " << pubEmptyFile << std::endl;

	const std::string contents = createEmptyFileContent(addressNum, thetime);

	fileWrite(pubEmptyFile, contents);
}



void setSingleAsEmptyAmount(const std::string& pubFolder, const int addressNum, const std::string& thetime)
{
	const std::string pubEmptyFile = SingleFolderHandler::getSinglePubEmptyFile(pubFolder, addressNum);

	std::cout << "Writing file: " << pubEmptyFile << std::endl;

	const std::string contents = createEmptyFileContent(addressNum, thetime);

	fileWrite(pubEmptyFile, contents);
}



void setPrivAsEmptyAmount(const std::list<std::string>& folders, const int addressNum, const std::string& thetime)
{
	int padnum(0);
	for(const std::string& privFolder : folders)
	{
		padnum++;		
		const std::string privEmptyFile = OtpFolderHandler::getPrivEmptyFile(privFolder, padnum, addressNum);
		
		if(privEmptyFile.empty())
		{
			continue;
		}
		
		std::cout << "Writing file: " << privEmptyFile << std::endl;
		const std::string contents = createEmptyFileContent(addressNum, thetime);
		fileWrite(privEmptyFile, contents);
		
	}

}


void setPrivFolderAsEmptyAmount(const std::string& folder, const int addressNum, const std::string& thetime)
{
	const std::string privEmptyFile = SingleFolderHandler::getSinglePrivEmptyFile(folder, addressNum);

	std::cout << "Writing file: " << privEmptyFile << std::endl;
	const std::string contents = createEmptyFileContent(addressNum, thetime);
	fileWrite(privEmptyFile, contents);
}




void GenerationMethodsOtp::markOtpAsEmptyAmount(const std::vector<std::string>& args, const std::string& thetime)
{	
	if(args.size() < 2)
	{
		std::cout << "Usage: bitgen otpempty NUMADDRESS" << std::endl;
		
		throw std::string("Incorrect usage");
	}

	const std::string numAddressStr = args[1];
	const int addressNum = atoi(numAddressStr.c_str());
	
	std::cout << "Setting as spent with amount zero for address: " << addressNum << std::endl;

	std::string publicFolder;
	const std::list<std::string> folders = BitgenConfig::getPublicAndOtpFolders(publicFolder);

	setAsEmptyAmount(publicFolder, addressNum, thetime);
	setPrivAsEmptyAmount(folders, addressNum, thetime);	
}



void GenerationMethodsOtp::markAsEmptyAmount(const std::vector<std::string>& args, const std::string& thetime)
{	
	if(args.size() < 2)
	{
		std::cout << "Usage: bitgen empty NUMADDRESS" << std::endl;
		
		throw std::string("Incorrect usage");
	}

	const std::string numAddressStr = args[1];
	const int addressNum = atoi(numAddressStr.c_str());
	
	std::cout << "Setting as spent with amount zero for address: " << addressNum << std::endl;

	std::string publicFolder;
	std::string privFolder;
	BitgenConfig::getThePublicAndPrivFolders(publicFolder, privFolder);

	setSingleAsEmptyAmount(publicFolder, addressNum, thetime);
	setPrivFolderAsEmptyAmount(privFolder, addressNum, thetime);	
}


void GenerationMethodsOtp::listSingleFolder(const std::vector<std::string>& args, const std::string& thetime)
{
	if(args.size() < 2)
	{
		std::cout << "Usage: bitgen list FOLDER" << std::endl;
		std::cout << "FOLDER: pub, priv" << std::endl;
		
		throw std::string("Incorrect usage");
	}

	const std::string itemStr = args[1];
	const int listItem = parsePubPriv(itemStr);

	int paramStartPos = 1;
	bool listAll(false);

	if(args.size() > 2)
	{
		const std::string allStr = args[2];
		if(allStr != "all")
		{
			std::string errStr("Unexpected argument:_");
			errStr += allStr;
			throw errStr;
		}
		
		paramStartPos++;
		listAll = true;
	}
	
	if(listAll)
	{
		std::cout << "Will list all existing addresses, including empty amounts" << std::endl;
	}
	

	GenerationParameters params;
	const std::vector<std::string> allExtraArgs = getExtraArgs(args, paramStartPos);	
	params.parseArgs(allExtraArgs);	
	params.finally();
	
	std::string publicFolder;
	std::string privFolder;
	BitgenConfig::getThePublicAndPrivFolders(publicFolder, privFolder);


	if(listItem == -1)
	{
		if(!directoryExist(publicFolder))
		{
			std::cout << "PUblic directory does not exist: " << publicFolder << std::endl;
			throw std::string("Folder does not exist");
		}

		InformationList::listPublicFolderContents(publicFolder, thetime, params.getAccountNameInternal(), listAll);			
	}
	else if(listItem == -2)
	{		
		if(!directoryExist(privFolder))
		{
			std::cout << "Private directory does not exist: " << privFolder << std::endl;
			throw std::string("Folder does not exist");
		}

		InformationList::listPrivateFolderContents(privFolder, thetime, params.getAccountNameInternal(), listAll);			
	}	
}





void GenerationMethodsOtp::listOtpFolder(const std::vector<std::string>& args, const std::string& thetime)
{
	if(args.size() < 2)
	{
		std::cout << "Usage: bitgen otplist FOLDER" << std::endl;
		std::cout << "FOLDER: pub, 1, 2 ... n" << std::endl;
		
		throw std::string("Incorrect usage");
	}

	const std::string numPadStr = args[1];
	const int numPad = parseNumpad(numPadStr);

	int paramStartPos = 1;
	bool listAll(false);

	if(args.size() > 2)
	{
		const std::string allStr = args[2];
		if(allStr != "all")
		{
			std::string errStr("Unexpected argument:_");
			errStr += allStr;
			throw errStr;
		}
		
		paramStartPos++;
		listAll = true;
	}
	
	if(listAll)
	{
		std::cout << "Will list all existing addresses, including empty amounts" << std::endl;
	}
	

	GenerationParameters params;
	const std::vector<std::string> allExtraArgs = getExtraArgs(args, paramStartPos);	
	params.parseArgs(allExtraArgs);	
	params.finally();
	
	std::string publicFolder;
	const std::list<std::string> folders = BitgenConfig::getPublicAndOtpFolders(publicFolder);

	if(numPad == -1)
	{
		if(!directoryExist(publicFolder))
		{
			std::cout << "PUblic directory does not exist: " << publicFolder << std::endl;
			throw std::string("Folder does not exist");
		}

		InformationList::listPublicFolderContents(publicFolder, thetime, params.getAccountNameInternal(), listAll);			
	}
	else
	{
		BitgenConfig::checkGivenFolderExist(folders, numPad);
	
		InformationList::generateOtpListForPad(folders, numPad, thetime, params.getAccountNameInternal(), listAll);
	}
}















void GenerationMethodsOtp::generateOtpWithPads(const std::list<BigInt<256> >& pads, const std::string method, 
	const std::list<std::string>& folders, const std::string& publicFolder,
	bool useConfiguredFolderPaths,
	std::string extraVal, const std::string& thetime, GenerationParameters& params)
{
	BigInt<256> privKey;
	for(std::list<BigInt<256> >::const_iterator it = pads.begin() ;
		it != pads.end() ;
		++it)
	{
		const BigInt<256>& pad = (*it);
		std::cout << "Pad: " << hexString(pad.getBinaryLong()) << std::endl;
		
		privKey = privKey ^ pad;
	}

	//Check if a static pad is defined
	const BigInt<256> staticXor = BitgenConfig::getStaticXor();
	//std::cout << "Static xor: " << hexString(staticXor.getBinaryLong()) << std::endl;
	
	const BigInt<256> zero;
	
	const bool useStaticPad = (staticXor != zero);
	
	std::list<ExtraInfo> generationExtra;
	
	if(useStaticPad)
	{
		const std::string hash = doubleSha256FourBytes(staticXor);		
		
		std::cout << "Using static pad: " <<  hexString(staticXor.getBinaryLong()) << " (" <<
			hexString(hash) << ")" << std::endl;
			
		const ExtraInfo extraInfoStaticPad("Static pad", hexString(hash));	
		generationExtra.push_back(extraInfoStaticPad);
	}


	privKey = privKey ^ staticXor;
	
	
	std::cout << "Privkey: " << hexString(privKey.getBinaryLong()) << std::endl;
		
	const PrivKey myPrivKey(privKey, true);	

		
	const std::string hexStr = hexString(privKey.getBinaryLong());

	BitcoinEllipticCurve ec;

	const Wif wif = myPrivKey.getWif();
	const PubKey pubKey = myPrivKey.getPubkey(ec);
	const PubAddress pubAddr = pubKey.getAddress();
	
	
	std::cout << "PUB : " << pubAddr << std::endl;		
	std::cout << "PRIV: " << wif.toString() << std::endl;
		

	const std::string generator = ((method == "random") ? "OTP /dev/random" : ("OTP " + method));
	
	std::cout << std::endl;

	std::vector<std::string> privatePaths;	
	std::string pubPath;

	const int newNum = SaveFolderHandler::getNextNumber();
	
	if(!folders.empty())
	{
		const std::string firstFolder = folders.front();

		const std::pair<std::string, std::vector<std::string> > fullFolders = 
			SaveFolderHandler::createNewFoldersOtp(newNum, folders, publicFolder, true, true);

		privatePaths = fullFolders.second;
		pubPath = fullFolders.first;
	}
	else
	{
		for(int i = 0 ; i < pads.size() ; i++)
		{
			privatePaths.push_back(".");
		}
		
		pubPath = ".";
	}

	std::cout << std::endl;

	const std::list<std::string> hashParts = splitHashStr(pads.size(), extraVal);

	const GenerationInfoBase infoBase(thetime, getNodeName(), applicationName, versionNumber);		
	const GenerationInfoCore infoCore(infoBase, newNum, params);

	GenerationInfo info(infoCore, generationExtra, generator);
	
	info.setStaticXor(staticXor);

	
	GenerationMethodsOtp::savePrivatePadPart(info, pads, hashParts, privatePaths, pads.size());

	std::cout << std::endl;

	const AddressGenerationInfo allInfo(pubAddr.toString(), "DummyPrivAddr", "Dummy", info);

	GenerationMethodsOtp::savePublicPadPart(allInfo, pubPath);
	

	if(useStaticPad)
	{
		std::cout << std::endl;
		std::cout << "NOTE: STATIC PAD USED. THE PAD MUST BE SECURELY STORED" << std::endl;
		std::cout << "Without knowledge of the static pad, the private key can not be recovered" << std::endl;
	}	
}







void GenerationMethodsOtp::generateOtpParts(const std::vector<std::string>& args, const std::string& thetime)
{
	if(args.size() < 3)
	{
		std::cout << "Usage: bitgen otpgen [NUMPAD/useconf] METHOD [extraOptions]" << std::endl;
		std::cout << "  Methods: random, dice, hash, hex" << std::endl;
		
		throw std::string("Incorrect usage");
	}
	
	const std::string numStr = args[1];
	std::list<std::string> folders;
	
	int totalPads(0);
	std::string publicFolder;
	
	if(numStr == "useconf")
	{
		std::cout << "Using configured folders for num pads" << std::endl;
		folders = BitgenConfig::getPublicAndOtpFolders(publicFolder);		
		BitgenConfig::checkFoldersExist(folders);
		std::cout << "OK, configured folders exist" << std::endl;
		
		totalPads = folders.size();
	}
	else
	{
		totalPads = atoi(numStr.c_str());
		
		//for(int i = 0 ; i < totalPads ; i++)
		//{
		//	folders.push_back(".");
		//}
	}	
	
	if(totalPads < 2)
	{
		std::cout << "Num pads given: " << totalPads << " (" << numStr << ")" << std::endl;
		std::cout << "Num pads must be at least 2" << std::endl;
		throw std::string("Incorrect number of pads");
	}
	
	std::cout << "Total pads: " << totalPads << std::endl;

	const std::string method = args[2];
	
	if(method != "random" && method != "hash" && method != "dice" && method != "hex" && method != "raw")
	{
		std::cout << "Unknown method: " << method << std::endl;
		throw std::string("Unknown method");
	}
	
	std::string extraVal;


	if(method == "random")
	{
		//Nothing
	}
	else if(method == "hash")
	{
		if(args.size() < 4)
		{
			std::cout << "Usage: bitgen otpgen NUMPAD hash HASHVAL [extraOptions]" << std::endl;
		
			throw std::string("Incorrect usage");
		}
		
		extraVal = args[3];		//Todo: Rename to paramVal
	}
	else if(method == "dice")
	{
		if(args.size() < 4)
		{
			std::cout << "Usage: bitgen otpgen NUMPAD dice DICEVAL [extraOptions]" << std::endl;
		
			throw std::string("Incorrect usage");
		}
		
		extraVal = args[3];		//Todo: Rename to paramVal
	}
	else if(method == "hex")
	{
		if(args.size() < 4)
		{
			std::cout << "Usage: bitgen otpgen NUMPAD hex HEXVAL [extraOptions]" << std::endl;
		
			throw std::string("Incorrect usage");
		}
		
		extraVal = args[3];		//Todo: Rename to paramVal
	}
	else if(method == "raw")
	{
		if(args.size() < 4 || (args.size() < (3 + totalPads)))
		{
			std::cout << "Usage: bitgen otpgen NUMPAD/useconf raw HEX1 HEX2 ... HEXN [extraOptions]" << std::endl;
		
			throw std::string("Incorrect usage");
		}

		extraVal = "raw";

		std::list<BigInt<256> > pads;
		
		for(int i = 0 ; i < totalPads ; i++)
		{
			const std::string rawHex = args[3 + i];
			
			if(rawHex.size() != 64)
			{
				std::cout << "Error, size should be 64, got " << rawHex.size() << " :" << rawHex << std::endl;
				throw std::string("Wrong size for raw pad");
			}
			
			const std::string binStr = convertFromHex(rawHex);
			const std::string hexValUpper = hexString(binStr);		
			const BigInt<256> pad = BigInt<256>::fromHex(hexValUpper);
			
			pads.push_back(pad);
		}

		GenerationParameters params;

		const int paramStartPos = 2 + totalPads;
		
		const std::vector<std::string> allExtraArgs = getExtraArgs(args, paramStartPos);
		params.parseArgs(allExtraArgs);		
		params.finally();
		
		generateOtpWithPads(pads, method, folders, publicFolder, (numStr != "useconf"), extraVal, thetime, params);		
		return;
	}
	else
	{
		std::cout << "Internal error, no such method: " << method << std::endl;
		throw std::string("No such method");
	}
	
	GenerationParameters params;

	const int paramStartPos = (method == "random") ? 2 : 3;

	const std::vector<std::string> allExtraArgs = getExtraArgs(args, paramStartPos);
	params.parseArgs(allExtraArgs);		
	params.finally();

	
	bool useFolder = (numStr != "useconf") ? false : params.getUseFolder();
	bool useConfiguredFolderPaths(false);
	
	if(numStr == "useconf")		//Using folder conf implies that folders are used
	{
		useFolder = true;
		useConfiguredFolderPaths = true;
	}

	const std::list<BigInt<256> > pads = generatePadRandom(totalPads, method, extraVal);


	std::cout << std::endl;

	generateOtpWithPads(pads, method, folders, publicFolder, useConfiguredFolderPaths, extraVal, thetime, params);
}








