/**
 * bitgencli.cpp - Bitcoin address generator
 *
 * Copyright (C)2015 The bitgen developers <bitgen@openmailbox.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
 
/**
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE. 
*/
 



#include "test/unittest.h"
#include "privKey.h"
#include "pubKey.h"
#include "bitgenVersion.h"

#include "commandLine.h"
#include "keyDerivationFunction.h"
#include "txtGeneration.h"
#include "currencySpecific.h"
#include "license.h"
#include "GenerationParameters.h"
#include "usagePrintout.h"
#include "GenerationMethods.h"
#include "bitlibVersion.h"
#include "CommandlineParser.h"
#include "MiscCommandline.h"
#include "FileUtil.h"
#include "bitgenConfig.h"
#include "currencySpecific.h"
#include "language.h"

#include <time.h>
#include <iostream>

#ifndef MINIMAL_GENERATION
#include "miniPriv.h"
#include "RangeSearcher.h"
#include "HierarchialRandom.h"
#include "StaticXor.h"
#include "MnemonicCommandline.h"
#include "ExtendedAddressCommandline.h"
#include "TransactionCommandline.h"
#include "RootGenerationCommandline.h"
#include "EntityInformationCommandline.h"
#include "ElectrumCommandline.h"
#endif


void printCompleteVersion()
{
	std::cout << "bitgen version " << versionNumber << std::endl;
	std::cout << std::endl;
	std::cout << "using bigint " << BigInt<0>::version() << std::endl;
	std::cout << "using bitlib version " << bitlibVersionNumber << std::endl;
	std::cout << std::endl;

	std::cout << "Currency: " << getCurrencyName() << " (" << getCurrencyShortName() << ")" << std::endl;
	std::cout << std::endl;
		
	std::cout << "http://www.bitgen.org/" << std::endl;		
}










void removeOtpFolders(const std::vector<std::string>& args)
{
	BitgenConfig::removeFolderConfig();
}


void removeSingleFolders(const std::vector<std::string>& args)
{
	BitgenConfig::removeSingleFolderConfig();
}


void setNextNumToUse(const std::vector<std::string>& args)
{
	if(args.size() < 2)
	{
		std::cout << "Usage: bitgen setnextnum NUMBER" << std::endl;
		return;
	}
	
	const std::string numStr = args[1];
	
	const int num = atoi(numStr.c_str());
	
	std::cout << "Setting next number to use: " << num << std::endl;
	
	BitgenConfig::setConfiguredNum(num);
}



std::string canonFileName(const std::string& folder)
{
#ifdef __linux
	char* absolutePathRaw = realpath(folder.c_str(), NULL);
	if(absolutePathRaw == NULL)
	{
		std::string errMsg("Folder does not exist: ");
		errMsg += folder;
		
		std::cout << errMsg << std::endl;
		
		throw errMsg;
	}
	
	const std::string absPath(absolutePathRaw);
	free(absolutePathRaw);
	
	return absPath;
#else
	throw std::string("realpath is not supported on windows");
	return "not_supported_on_windows";	
#endif
}


void setSingleFolders(const std::vector<std::string>& args)
{
	if(args.size() < 2)
	{
		std::cout << "Usage: bitgen otp PUBFOLDER PRIVFOLDER" << std::endl;
		
		std::cout << std::endl;
	
		std::string confPublicFolder;
		std::string privFolder;
		BitgenConfig::getThePublicAndPrivFolders(confPublicFolder, privFolder);
		
		std::cout << "Configured folders: " << std::endl;
		std::cout << "Public: " << confPublicFolder << std::endl;
		std::cout << "Private: " << privFolder << std::endl;

		return;
	}	

	std::cout << "Setting Single folders" << std::endl;
	
	std::list<std::string> folders;
	
	const std::string pubFolder = args[1];
	const std::string privFolder = args[2];
	
	std::cout << "Pub : " << pubFolder << std::endl;
	std::cout << "Priv: " << privFolder << std::endl;
	
	BitgenConfig::addSingleFoldersToConfig(pubFolder, privFolder);	
}


void setOtpFolders(const std::vector<std::string>& args)
{
	if(args.size() < 2)
	{
		std::cout << "Usage: bitgen otpdir PUBFOLDER DIR1 ... DIRN" << std::endl;
		
		std::cout << std::endl;
	
		std::string confPublicFolder;
		const std::list<std::string> folders = BitgenConfig::getPublicAndOtpFolders(confPublicFolder);
		
		std::cout << "Configured folders: " << std::endl;
		std::cout << "Public: " << confPublicFolder << std::endl;
		std::cout << std::endl;
		for(std::list<std::string>::const_iterator it = folders.begin() ;
			it != folders.end() ;
			++it)
		{
			const std::string& folder = (*it);
			std::cout << folder << std::endl;
		}
		
		return;
	}	


	std::cout << "Setting OTP folders" << std::endl;
	
	std::list<std::string> folders;
	for(int i = 1 ; i < args.size() ; i++)
	{
		const std::string canonName = canonFileName(args[i]);
		folders.push_back(canonName);
	}
	
	BitgenConfig::addFoldersToConfig(folders);
	
	for(std::list<std::string>::const_iterator it = folders.begin() ;
		it != folders.end() ;
		++it)
	{
		const std::string& folder = (*it);
		std::cout << "Folder: " << folder << std::endl;
	}
}


void addName(const std::vector<std::string>& args, const std::string& thetime)
{

	if(args.size() < 2)
	{
		std::cout << "Usage: bitgen addname NAME" << std::endl;
		throw std::string("Not enough arguments");
	}

	const std::string name = args[1];

	std::cout << "Adding name: " << name << std::endl;
	BitgenConfig::configureName(name);
}


void removeName(const std::vector<std::string>& args, const std::string& thetime)
{	
	bool hasNameConf(false);
	const std::string currentName = BitgenConfig::getConfiguredName(hasNameConf);
	
	if(hasNameConf)
	{
		std::cout << "Removing name: " << currentName << std::endl;
		BitgenConfig::deleteName();		
	}
	else
	{
		std::cout << "No existing name to remove" << std::endl;
	}	
}



void getNextNumToUse()
{
	const int nextNum = BitgenConfig::getConfiguredNum(false);
	
	std::cout << "nextNum: " << nextNum << std::endl;
}



bool dispatchSettings(const std::string& firstArg, const std::vector<std::string>& args, const std::string& thetime)
{
	if(firstArg == "dir")
	{
		setSingleFolders(args);
		return true;
	}
	else if(firstArg == "removedir")
	{
		removeSingleFolders(args);
		return true;
	}	
	else if(firstArg == "otpdir")
	{
		setOtpFolders(args);
		return true;		
	}
	else if(firstArg == "removeotpdir")
	{
		removeOtpFolders(args);
		return true;
	}						
	else if(firstArg == "setlang")
	{
		setLanguage(args, thetime);
		return true;
	}		
	else if(firstArg == "getnextnum")
	{
		std::cout << "Getting next num" << std::endl;
		getNextNumToUse();
		return true;
	}
	else if(firstArg == "setnextnum")
	{		
		std::cout << "Setting next num" << std::endl;
		setNextNumToUse(args);
		return true;
	}
	else if(firstArg == "addname")
	{
		addName(args, thetime);
		return true;
	}	
	else if(firstArg == "removename")
	{
		removeName(args, thetime);
		return true;
	}
	else
	{
		return false;
	}
}




#ifndef MINIMAL_GENERATION


bool dispatchOtp(const std::string& firstArg, const std::vector<std::string>& args, const std::string& thetime)
{		
	//One time pad
	if(firstArg == "otpgen")
	{
		GenerationMethodsOtp::generateOtpParts(args, thetime);
		return true;
	}
	else if(firstArg == "otp")
	{
		GenerationMethodsOtp::generateOtpTotal(args, thetime);
		return true;
	}	
	else if(firstArg == "list")
	{
		GenerationMethodsOtp::listSingleFolder(args, thetime);
		return true;
	}
	else if(firstArg == "empty")
	{
		GenerationMethodsOtp::markAsEmptyAmount(args, thetime);
		return true;
	}	
	else if(firstArg == "otplist")
	{
		GenerationMethodsOtp::listOtpFolder(args, thetime);
		return true;
	}	
	else if(firstArg == "otpempty")
	{
		GenerationMethodsOtp::markOtpAsEmptyAmount(args, thetime);
		return true;
	}
	//Static xor
	else if(firstArg == "addxor")
	{
		addStaticXor(args, thetime);
		return true;		
	}	
	else if(firstArg == "removexor")
	{
		removeStaticXor(args, thetime);
		return true;		
	}	
	else if(firstArg == "readxor")
	{
		readStaticXor(args, thetime);
		return true;		
	}	
	else
	{
		return false;
	}
}




bool dispatchTransaction(const std::string& firstArg, const std::vector<std::string>& args, const std::string& thetime)
{
	if(firstArg == "savetrans")
	{
		TransactionCommandline::saveTheTransaction(args);
		return true;
	}	
	else if(firstArg == "decodetrans")
	{
		TransactionCommandline::decodeTransaction(args);
		return true;		
	}
	else if(firstArg == "parsetrans")		//Redundant?
	{
		TransactionCommandline::parseTransaction(args);
		return true;		
	}
	else if(firstArg == "transinputhashes")	//Redundant?
	{
		TransactionCommandline::getTransactionInputHashes(args);
		return true;		
	}	
	else if(firstArg == "savescript")
	{
		TransactionCommandline::saveScript(args);
		return true;		
	}	
	else if(firstArg == "decodescript")
	{
		TransactionCommandline::decodeScript(args);
		return true;		
	}
	else if(firstArg == "runtrans")
	{
		TransactionCommandline::runTrans(args);
		return true;		
	}
	else if(firstArg == "transaction")
	{
		TransactionCommandline::createTransaction(args);
		return true;
	}
	else if(firstArg == "multisig")
	{
		createMultisig(args);
		return true;		
	}
	else if(firstArg == "signmultisig")	
	{
		partSignMultisig(args);
		return true;		
	}
	else if(firstArg == "combinemultisig")
	{
		combineMultisigSignatures(args);
		return true;		
	}	
	else
	{
		return false;
	}
}




bool dispatchExtended(const std::string& firstArg, const std::vector<std::string>& args, const std::string& thetime)
{
	//Extended keys
	if(firstArg == "genMnemonic")	//Deprecated, use "seed" below
	{
		std::cout << "Command \"genMnemonic\" is deprecated. Use \"seed\" instead" << std::endl;
		createMnemonicAccount(args, thetime);	//Deprecated
		return true;
	}
	else if(firstArg == "seed")
	{
		createMnemonicAccount(args, thetime);
		return true;
	}	
	else if(firstArg == "hmacsha512")
	{
		std::cout << "Moved to unit tests" << std::endl;
		return true;		
	}	
	else if(firstArg == "derivation")
	{
		calcDerivation(args);
		return true;		
	}
	else if(firstArg == "addressderivation")
	{
		calcAddressDerivation(args);
		return true;		
	}
	else if(firstArg == "firstaddress")
	{
		calcFirstHdAddress(args);
		return true;		
	}	
	else
	{
		return false;
	}
}	





bool dispatchHierarcial(const std::string& firstArg, const std::vector<std::string>& args, const std::string& thetime)
{
	//Commands for Hierarcial random addresses
	if(firstArg == "genRoot")
	{
		HierarcialRandomCommandline::generateRootKey(args, thetime);
		return true;
	}
	else if(firstArg == "genPub")
	{
		HierarcialRandomCommandline::generatePublicFromRoot(args, thetime);
		return true;		
	}
	else if(firstArg == "genPriv")
	{
		HierarcialRandomCommandline::generatePrivateFromRoot(args, thetime);
		return true;
	}
	else
	{
		return false;
	}
}









bool dispatchDeprecated(const std::string& firstArg, const std::vector<std::string>& args, const std::string& thetime)
{	
	if(firstArg == "mini")
	{
		miniKeyGenerate(args, thetime);
		return true;
	}
	else if(firstArg == "miniToWif")
	{
		generateWifFromMiniPrivKey(args, thetime);
		return true;
	}
	else if(firstArg == "verifyWIF") //Todo: Remove??
	{
		std::cout << "This command is deprecated. Use \"info\" instead" << std::endl;		
		std::cout << std::endl;		
		EntityInformationCommandline::verifyWif(args, thetime);
		return true;
	}			
	//Commands for public keys	
	else if(firstArg == "uncompress")	//Uncompresses the public key in hex format
	{
		uncompressPublicKey(args, thetime);
		return true;
	}
	else if(firstArg == "compress")		//Compresses the public key in hex format
	{
		compressPublicKey(args, thetime);
		return true;
	}
	else if(firstArg == "intsize")
	{
		printIntSize();
		return true;		
	}	
	else
	{
		return false;
	}
}



#endif




void doBitgenWork(int argc, char* argv[])
{			
	time_t t = time(0);
	const std::string thetime = removeNewline(asctime(localtime(&t)));
	
	std::vector<std::string> args;
	for(int i = 1 ; i < argc ; i++)
	{
		args.push_back(argv[i]);		
	}

	if(argc < 2)
	{
		std::cout << "For help, use \"bitgen-" << getCurrencyShortName() << " help\"" << std::endl;
		throw std::string("Missing argument");
		return;
	}

	const std::string firstArg(argv[1]);


	if(firstArg == "version" || firstArg == "-v")
	{
		printCompleteVersion();
	}
	else if(firstArg == "help")
	{
		printUsage();	
	}
	else if(firstArg == "license")
	{
		printLicense();
	}
	else if(firstArg == "single")
	{
		std::cout << "To be implemented" << std::endl;
		GenerationMethods::generateSingle(args, thetime);
	}	
	else if(firstArg == "brainwallet")
	{
		GenerationMethods::generateBrainWallet(args, thetime);
	}
	//Private key generation
	else if(firstArg == "bulk")
	{
		GenerationMethods::generateBulk(args, thetime);
	}	
	else if(firstArg == "dice")
	{
		GenerationMethods::generateWithDice(args, thetime);
	}
	else if(firstArg == "hash")
	{
		GenerationMethods::generateHash(args, thetime);	
	}
	else if(firstArg == "hex")
	{
		GenerationMethods::generateHex(args, thetime);
	}
	else if(firstArg == "mnemonic")		//Todo: remove
	{
		std::cout << "Mnemonic not supported" << std::endl;
		//GenerationMethods::generateFromMnemonic(args, thetime);	//Todo: remove	
	}		
	else if(firstArg == "random")
	{
		GenerationMethods::generateRandom(args, thetime);
	}
	//Raw random number generation
	else if(firstArg == "binary")
	{
		GenerationMethods::generateBinaryRandomNumber(args, thetime);
	}	
	else if(firstArg == "wif")
	{
		GenerationMethods::generateFromWif(args, thetime);
	}
	else if(dispatchSettings(firstArg, args, thetime))
	{
		
	}		

#ifndef MINIMAL_GENERATION
	else if(firstArg == "info")
	{
		EntityInformationCommandline::informationAboutEntity(args, thetime);
	}	
			
	//Commands that display a public address
	else if(firstArg == "pub")
	{
		GenerationMethods::generatePublicPart(args, argc, argv, thetime);
	}
	else if(firstArg == "invoice")
	{
		generatePaymentRequest(argc, argv, thetime);
	}
	
	//Electrum	
	else if(firstArg == "electrum")
	{
		createElectrumAccount(args, thetime);
	}
	//Extra
	
	
	else if(dispatchTransaction(firstArg, args, thetime))
	{
		//Nothing		
	}	
	else if(dispatchExtended(firstArg, args, thetime))
	{
		
	}		
	else if(dispatchOtp(firstArg, args, thetime))
	{
		//Nothing		
	}
	else if(dispatchHierarcial(firstArg, args, thetime))
	{

	}	
	else if(dispatchDeprecated(firstArg, args, thetime))
	{

	}
#endif

	else
	{
		std::cerr << "Unknown command: " << argv[1] << std::endl;
	}
	
}







